class ExportsController < ApplicationController
  require 'csv'

  include ArchiveHelpers

  class InvalidMachineTarget < StandardError; end
  class MissingDate < StandardError; end
  class InvalidStartDate < StandardError; end

  def get_jobs
    machine, start_date, end_date = export_params

    models = get_models(JobState, :machine_id, machine.id, start_date, end_date).where(state: 'finished').eager_load(:job)

    base_col_extract = lambda do |job_state|
      return [job_state.job.number, job_state.side, job_state.time]
    end

    row_data, dynamic_col_headers = extract_dynamic_row_data(models, base_col_extract, :process_params)

    @headers = ['Job Number', 'Side', 'Date - Time'] + dynamic_col_headers
    @rows = build_rows_from_dynamic_data(row_data, dynamic_col_headers)

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_machine_states
    machine, start_date, end_date = export_params

    @headers = ['State', 'Date - Time']
    @rows = []

    get_models(MachineState, :machine_sn, machine.sn, start_date, end_date)
            .find_each do |machine_state|
              @rows << [machine_state.state, machine_state.time]
            end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_job_states
    machine, start_date, end_date = export_params

    @headers = ['Job Number', 'Side', 'State', 'Substate', 'Scrap', 'Scrap Reason', 'Comment', 'Date - Time']
    @rows = []

    get_models(JobState, :machine_id, machine.id, start_date, end_date).eager_load(:job)
            .find_each do |job_state|
              job_scrap = if job_state.scrap.nil?
                            'F'
                          else
                            job_state.scrap ? 'T' : 'F'
                          end
              job_number = if job_state.job.nil?
                             'n/a'
                           else
                             job_state.job.number
                           end

              @rows << [job_number, job_state.side, job_state.state, job_state.substate, job_scrap, job_state.scrap_reason, job_state.comment, job_state.time]
            end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_errors
    machine, start_date, end_date = export_params

    @headers = ['Error', 'Err No.', 'Error Type', 'Status', 'Date - Time']
    @rows = []

    get_models(Error, :machine_sn, machine.sn, start_date, end_date)
            .find_each do |error|
              @rows << [error.error, error.errno, error.errortype, error.status, error.time]
            end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_events
    machine, start_date, end_date = export_params

    models = get_models(Event, :machine_sn, machine.sn, start_date, end_date)

    base_col_extract = lambda do |event|
      return [event.message, event.event_type, event.time]
    end

    row_data, dynamic_col_headers = extract_dynamic_row_data(models, base_col_extract, :event_data)

    @headers = ['Message', 'Type', 'Date - Time'] + dynamic_col_headers
    @rows = build_rows_from_dynamic_data(row_data, dynamic_col_headers)

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_maintenances
    machine, start_date, end_date = export_params

    @headers = ['Maintenance', 'Maintenance Type', 'Status', 'Date - Time']
    @rows = []

    get_models(Maintenance, :machine_sn, machine.sn, start_date, end_date)
            .find_each do |maintenance|
              @rows << [maintenance.maintenance, maintenance.maintenance_type, maintenance.status, maintenance.time]
            end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_calibrations
    machine, start_date, end_date = export_params

    @headers = ['Target', 'Target Name', 'Calibration Type', 'Sub Type', 'Mode', 'User', 'Date - Time']
    @rows = []

    get_models(Calibration, :machine_sn, machine.sn, start_date, end_date)
            .find_each do |calibration|
              @rows << [calibration.target, calibration.target_name, calibration.calibration_type, calibration.sub_type, calibration.mode, calibration.user, calibration.time]
            end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  def get_process_data
    machine, start_date, end_date = export_params

    @headers = ['Job Number', 'Side', 'Date - Time', 'Tag', 'Value']
    @rows = []

    JobProcessParameter.joins(:job_process).where(job_process: { time: start_date..end_date, machine: machine }).eager_load(job_process: :job)
              .find_each do |job_process_parameter|
                next if job_process_parameter.job_process.nil?
                job = job_process_parameter.job_process.job
                next if job.nil?

                @rows << [job.number, job.side, job_process_parameter.job_process.time, job_process_parameter.name, job_process_parameter.value]
              end

    render json: { csv: build_export_csv }, status: :ok
  rescue => e
    render json: { error: e }, status: :unprocessable_entity
  end

  private

  def export_params
    machine_id = params[:machine_id]
    start_date_str = params[:start_date]
    end_date_str = params[:end_date]

    start_date, end_date = validate_dates(start_date_str, end_date_str)
    machine = machine_target(machine_id)

    [
      machine,
      start_date,
      end_date
    ]
  end

  def validate_dates(start_date, end_date)
    raise MissingDate if start_date.nil? || end_date.nil?

    time_start_date, time_end_date = start_date.to_time, end_date.to_time

    raise InvalidStartDate if time_start_date > time_end_date

    [time_start_date, time_end_date]
  end

  def machine_target(machine_id)
    machine = Machine.find_by(id: machine_id)

    raise InvalidMachineTarget if machine.nil?

    machine
  end

  def build_export_csv
    csv_string = CSV.generate do |csv|
      csv << @headers

      @rows.each do |row|
          csv << row
      end
    end
    csv_string
  end

  def get_models(model, machine_symbol, machine_value, start_date, end_date)
    shard = if time_exists_in_primary? start_date
              :primary
            else
              :archive
            end

    ActiveRecord::Base.connected_to(shard: shard, role: :reading) do
      model.where(
        model.arel_table[machine_symbol].eq(machine_value).and(
          model.arel_table[:time].gteq(start_date).and(model.arel_table[:time].lteq(end_date))
        )
      )
    end
  end

  # The columns of these row will depend on the row item themselves (for example, each process data adds a new column with the process name)
  def extract_dynamic_row_data(models, base_col_extract, col_association)
    col_headers = []
    row_data = []

    models.find_each do |model|
      col_values = {}

      # Fetch the association data (ex: event_data, process_params, etc.)
      association_data = model.send(col_association)

      unless association_data.blank?
        association_data.each do |data|
          col_headers << data.name
          col_values[data.name] = data.value
        end
      end

      row_data << {
        base_data: base_col_extract.call(model),
        dynamic_data: col_values
      }
    end

    [row_data, col_headers]
  end

  def build_rows_from_dynamic_data(row_data, dynamic_col_header)
    rows = []

    row_data.each do |data|
      row = data[:base_data]

      dynamic_col_header.each do |col|
        row << if data[:dynamic_data].key? col
                data[:dynamic_data][col]
               else
                 nil
               end
      end
      rows << row
    end

    rows
  end
end
