class OrdersController < ApplicationController
  include MachineResponse
  include OrderCreateHelper

  wrap_parameters :order, include: [:address, :shipTo, :data]
  model_data_store(Order)

  def add
    begin
      AddOrderJob.perform_later(params[:id], add_params, address_params, ship_to_params, data_params)
    rescue => e
      Rails.logger.error e
    end
    send_machine_response
  end

  def create_from_routing
    job_data = create_order_and_jobs(get_order_id_from_evt, add_params, address_params, ship_to_params, data_params, process_params)
    render json: job_data, status: :created
  end

  def add_order_and_jobs
    begin
      AddOrderJobsJob.perform_later(get_order_id_from_evt, add_params, address_params, ship_to_params, data_params, process_params, params.to_json)
    rescue => e
      Rails.logger.error e
    end
    send_machine_response
  end

  def revive
    @model = Order.find_by(id: params[:order_id])
    if @model.nil?
      head :unprocessable_entity
      return
    end
    @model.revive
    render :show
  end

  def delete
    # 2023/03/03 - Keeping this endpoint for legacy support. Do not deprecate as Rx.Net calls this.
    render html: "OK"
  end

  def bulk_order_data
    order_ids = bulk_data_order_params[:order_ids]

    render json: {error: "Invalid order_ids"}, status: :unprocessable_entity if order_ids.blank?

    order_data = {}

    order_ids.each do |order_id|
      order = Order.find_by(id: order_id)

      if order.nil?
        order_data[order_id] = {error: 'order_not_found'}
      else
        order_data[order_id.to_s] = {tags: order.routing_launch_data[:tag_data]}
      end
    end

    render json: order_data, status: :ok
  end

  # DEPRECATED LEGACY ENDPOINTS

  def types
    # DEPRECATED 2023/03/03 - #get /jobinfo/1.0/ordertype/list
    render json: {status: 'Endpoint Deprecated'}, status: :ok
  end

  def data
    # DEPRECATED 2023/03/03 - #get /jobinfo/1.0/order/data
    render json: {status: 'Endpoint Deprecated'}, status: :ok
  end

  # DEPRECATED 2023/08/14
  def jobs
    ActiveRecord::Base.connected_to(shard: :archive, role: :writing) do
      @limit = params[:limit] || 200
      @offset = params[:offset] || 0

      max_age = DataArchiveSetting.instance.job_max_age

      @jobs = if params[:orderNr].present?
                FindHelpers::OrderFinders.find_active_or_create_by_number(params[:orderNr])&.jobs
              elsif params[:trayId].present?
                Order.joins(:jobs).find_by(jobs: {number: params[:trayId]})&.jobs
              end
      if @jobs
        @jobs = @jobs.where(expiry_date: max_age.hours.ago..).limit(@limit).offset(@offset)
        render :jobs
      else
        head :not_found
      end
    end
  end

  private

  # Only allow a list of trusted parameters through.
  def model_params
    params.require(:order).permit(:job_number, :name, :description, :order_type, :workflow_state, :priority, :patient, :rtime, :wtime)
  end

  def add_params
    order = dig_to_order
    {
      name: order[:name],
      description: order[:description],
      order_type: order[:ordertype],
      workflow_state: order[:workflowstate],
      priority: order[:priority],
      patient: order[:patient],
      rtime: "#{order[:rdate]} #{order[:rtime]}",
      wtime: "#{order[:wdate]} #{order[:wtime]}"
    }
  end

  def permit_address(key)
    address = dig_to_order.dig(key)
    address.nil? ? {} :
    {
      country: address[:country],
      postal_code: address[:postcode],
      street: address[:street],
      city: address[:city],
      name: address[:name]
    }
  end

  def address_params
    permit_address(:address)
  end

  def ship_to_params
    permit_address(:shipTo)
  end

  def data_params
    data_obj = dig_to_order.dig(:data)
    if data_obj.is_a?(String) || data_obj.nil?
      return []
    end
    data = data_obj.permit(item: [:name, :value, :lensside]).dig(:item)
    Array(data).map do |datum|
      {
        name: datum[:name],
        value: datum[:value],
        side: Sideable.side_letter_to_side_num(datum[:lensside])
      }
    end
  end

  def process_params
    process_obj = dig_to_order.dig(:processes)
    if process_obj.is_a?(String) || process_obj.nil?
      return []
    end
    processes = process_obj.permit(
      process: [
        :name,
        :enabled,
        :eye,
        plannedQuantity: [:value],
        quantityProductionTime: [:value],
        setupTime: [:value],
        leadTime: [:value],
        capabilities: [
          capability: [:name],
          "capability-group": [:name]
        ]
      ]
    ).dig(:process)
    Array(processes).map do |process|
      # capabilities = process[:capabilities].map do |capability|
      #   cap = Capability.get_by_name(capability[:capability][:name])
      #   cap_group = CapabilityGroup.get_by_name(capability[:"capability-group"][:name])
      #   cap.capability_groups << cap_group
      #   cap
      # end
      {
        name: process[:name],
        eye: process[:eye],
        planned_quantity: process.dig(:plannedQuantity)&.dig(:value),
        quantity_production_time: process.dig(:quantityProductionTime)&.dig(:value),
        setup_time: process.dig(:setupTime)&.dig(:value),
        lead_time: process.dig(:leadTime)&.dig(:value),
        enabled: process[:enabled]
        # capabilities: capabilities # we're ignoring this until I can find out where it's supposed to go
      }
    end
  end

  def get_order_id_from_evt
    params.require(:mes).require(:evt).permit(:id)[:id]
  end

  def dig_to_order
    if params.has_key?(:mes)
      params.require(:mes).require(:evt).require(:order)
    else
      params.require(:order)
    end
  end

  def bulk_data_order_params
    params.permit(order_ids: [])
  end
end
