import * as React from "react";
import {Button, CircularProgress, Grid, Typography} from "@material-ui/core";
import {observer} from "mobx-react-lite";
import {makeStyles} from "@material-ui/core/styles";
import {MachineModel, getProductionData} from "../../stores/MachineStore";
import {MachinePerformanceReport} from "./prod_view/MachinePerformanceReport";
import {ButtonField, Conditional, IfFalse, IfTrue, TemplateReportLayout} from "@rx-suite/ui";
import {MachineErrorReport} from "./prod_view/MachineErrorReport";
import {MachineAbortReport} from "./prod_view/MachineAbortReport";
import {MachineTypeStore} from "../../stores/MachineTypeStore";
import {useMemo} from "react";
import RefreshIcon from "@material-ui/icons/Refresh";
import {t} from "@lingui/macro";
import {SystemSettingModel, SystemSettingStore} from "@rx-suite/stores";
import {BatchCoaterOverviewTimeline} from "./prod_view/BatchRunOverviewTimeLine";

type MachineProductionViewProps = {
    model: MachineModel
}

const useStyles = makeStyles((theme) => ({
    refreshButtonWrap: {
        width: 'max-content'
    }
}));

export const MachineProductionView: React.FunctionComponent<MachineProductionViewProps> = observer(( {model}) => {
    const systemSettingModel = SystemSettingStore.selectedModel as SystemSettingModel;
    const classes = useStyles();

    const [isLoading, setIsLoading] = React.useState<boolean>(false);
    const [refreshTime, setRefreshTime] = React.useState<string | null>(null);
    const [startTime, setStartTime] = React.useState<string | null>(null);
    const [endTime, setEndTime] = React.useState<string | null>(null);
    const [prodData, setProdData] = React.useState<any|null>(null);

    const getProdViewData = async() => {
        setIsLoading(true);
        setRefreshTime(null);

        const response = await getProductionData(model.getId());

        setRefreshTime(response.time);
        setStartTime(response.start_time);
        setEndTime(response.end_time);
        setProdData({
            performanceReport: response.data.find(c => c.meta.id == "machine_performance_overview" || "machine_performance_batch"),
            errorReport: response.data.find(c => c.meta.id == "machine_top_errors"),
            abortReport: response.data.find(c => c.meta.id == "machine_top_aborts")
        });

        setIsLoading(false);
    }

    React.useEffect(() => {
        getProdViewData();
    }, [model])

    const machineType = MachineTypeStore.get(model.machine_type_id)

    const performanceReportComponent = useMemo(() => {
        if(!prodData || !prodData.performanceReport){
            return null;
        }

        if(machineType && machineType.isBoxCoater){
            return (
                 <BatchCoaterOverviewTimeline data={prodData.performanceReport.data} startTime={startTime} endTime={endTime}/>
            )
        }

        return <MachinePerformanceReport component={prodData.performanceReport} />
    }, [prodData, machineType]);

    const formattedLastRefreshTime = useMemo(() => {
        if(!systemSettingModel || !refreshTime){
            return t`Loading...`;
        }

        return systemSettingModel.printDateTime(new Date(refreshTime))
    }, [systemSettingModel, refreshTime]);

    return (
        <div>
            <Grid container direction={'row-reverse'}>
                <Grid container item direction={'column'} alignItems={'flex-end'} className={classes.refreshButtonWrap}>
                    <div>
                        <ButtonField
                            startIcon={<RefreshIcon/>}
                            text={t`Refresh`}
                            onClick={getProdViewData}
                            disabled={isLoading}
                            variant='outlined'
                            color='secondary'
                        />
                    </div>
                    <Typography variant={'subtitle2'}>{t`Last Refresh: ${formattedLastRefreshTime}`}</Typography>
                </Grid>
            </Grid>
            <Conditional condition={isLoading}>
                <IfTrue>
                    <Grid container justifyContent={'center'}>
                        <CircularProgress color={'secondary'} size={100}/>
                    </Grid>
                </IfTrue>
                <IfFalse>
                    {prodData &&
                        <TemplateReportLayout
                            rows={[
                                {
                                    components: [
                                        {
                                            element: performanceReportComponent as JSX.Element,
                                            columnSize: 10
                                        },
                                    ],
                                    height: '250px'
                                },
                                {
                                    components: [
                                        {
                                            element: <MachineErrorReport component={prodData.errorReport}/>,
                                            columnSize: 5
                                        },
                                        {
                                            element: <MachineAbortReport component={prodData.abortReport}/>,
                                            columnSize: 5
                                        },
                                    ],
                                    height: '250px'
                                }
                            ]}
                        />
                    }
                </IfFalse>
            </Conditional>
        </div>
    )
})