import * as React from 'react';
import {useState} from "react";
import {observer} from "mobx-react-lite";
import {useSnackbar} from "notistack";
import {t} from "@lingui/macro";
import {
    Grid,
    makeStyles,
    Paper,
    Typography
} from "@material-ui/core";
import {WarningBox} from "../../components/generic/WarningBox";
import {TabsContainer} from "../../components/generic/TabsContainer";
import {MachineHistoryTable} from "../../components/job_history/MachineHistoryTable";
import {JobDataTable} from "../../components/job_history/JobDataTable";
import {JobProcessDataTable} from "../../components/job_history/JobProcessDataTable";
import {JobStateModel} from "../../stores/JobStateStore";
import {SelectField, sendEnqueueSnackbar, WorkstationGenericWrapper, WorkstationHelper} from "@rx-suite/ui";
import {OperatorScannerField} from "../../components/generic/OperatorScannerField";
import {SystemSettingModel, SystemSettingStore} from "@rx-suite/stores";

type JobOperatorFormProps = {

};

type ErrorKey = 'invalidJobNumber' | 'noResultsForJob' | 'cannotFetchData' | 'unknownError';

export enum Side {
    LEFT = 'left',
    RIGHT = 'right'
}

export const sideDisplayText = (side: Side, isBoxCoater: boolean | undefined) => {
    if (!isBoxCoater) {
        if (side === Side.LEFT) return t`Left`;
        if (side === Side.RIGHT) return t`Right`;
    }
    if (side === Side.LEFT) return t`Concave`;
    if (side === Side.RIGHT) return t`Convex`;
    throw new Error(`Invalid side: ${side}`);
};

type JobData = Readonly<{
    name: string | null;
    value: string | number | null;
    type: "tag" | "process" | 'base';
}>;

export type JobDataSides = Readonly<{
    [sideKey in Side]: ReadonlyArray<JobData> | null;
}>;

type JobState = Readonly<Partial<JobStateModel>>;

export type ProcessParams = Readonly<{
    name: string | null;
    value: string | null;
}>;

export type JobStateInfo = Readonly<{
    state: JobState;
    process_params: ReadonlyArray<ProcessParams> | null;
}>;

type JobStateSides = Readonly<{
    [sideKey in Side]: ReadonlyArray<JobStateInfo>;
}>;

export type MachineInfo = Readonly<{
    name: string;
    sn: string;
    description: string;
    states: JobStateSides;
    isBoxCoater: boolean; // change display text from left/right or concave/convex
}>;

export type MachineHistory = Readonly<{
    [machineKey: string]: MachineInfo;
}>;

type OrderUpdateTime = Readonly<{
    order_id: number,
    created_at: string
}>;

type AllOrdersUpdateTimes = Readonly<OrderUpdateTime[]>;

type JobHistoryResponse = {
    job_data: JobDataSides;
    machine_history: MachineHistory;
    all_orders_update_times: AllOrdersUpdateTimes;
    selected_order: number;
    // routing_history: {}; // TODO: Not implemented yet
};

const useStyles = makeStyles((theme) => ({
    searchBoxFull: {
        width: '100%'
    },
    results: {
        display: 'flex',
        flexDirection: 'column',
        marginTop: '0.5rem'
    },
    tabsContainer: {
        marginTop: '0.5rem'
    },
    chip: {
        padding: '0.5rem',
        borderBottom: '1px solid #80808017',
        display: 'flex'
    },
    orderSelector: {
        width: '15%',
        marginLeft: '5px'
    },
    createdAtText: {
        alignSelf: 'center',
        marginLeft: '15px'
    }
}));

// get the job history information related to a job number
const getJobByNumber = async (number: string, orderId:number|null = null) => {
    try {
        let params = `job_number=${number}`
        if(orderId) {
            params += `&order_id=${orderId}`
        }
        const response = await fetch(`/api/jobs/job_history?${params}`);
        const data = await response.json();
        return data;
    } catch (e) {
        // if the fetch fails, return an error object to the form
        return { error: 'cannotFetchData' };
    }
};

// get the process params (filtered, meaning no undefined) from the side of a machine
const getFilteredSideProcess = (machine: MachineInfo, side: Side): Array<Array<ProcessParams>> => {
    // fetch a copy of all the process params
    const sideProcess = machine.states[side].map(jobState => {
        const processParams = jobState.process_params;
        if (!processParams) return;
        return processParams.slice();
    });
    // removing undefined from process array
    // type assertion here is required, because filter returns ([] | undefined)
    // this means typescript will always think this is Array<Array<ProcessParams> | undefined>
    const filteredSideProcess = sideProcess.filter(p => !!p) as Array<Array<ProcessParams>>;
    return filteredSideProcess;
};

// fetch all process params (filtered and flattened)
const getProcessDataForSide = (machineHistory: MachineHistory, side: Side): ReadonlyArray<ProcessParams> => {
    const machines = Object.values(machineHistory);
    const processData = machines.map(machine => {
        // get the filtered process params
        const filteredProcessParams = getFilteredSideProcess(machine, side);
        // currently the array is in the form [[...], [...], ...]
        // the array needs to be flattened
        const flattenArr = filteredProcessParams.reduce((acc, val) => acc.concat(val), []);
        return flattenArr;
    });
    // currently the array is in the form [[...], [...], ...]
    // the array needs to be flattened
    const flattenedProcessData = processData.reduce((acc, val) => acc.concat(val), []);
    return flattenedProcessData;
};


export const JobHistoryOperatorForm: React.FunctionComponent<JobOperatorFormProps> = observer(() => {

    const {enqueueSnackbar} = useSnackbar();
    const classes = useStyles();
    const [jobInfo, setJobInfo] = useState<JobHistoryResponse|null>(null);
    const [jobNumber, setJobNumber] = useState<string|null>(null);
    const [orderValue, setOrderValue] = useState<number>(1);
    const [state, setState] = useState<'ready'|'loading'|'error'>('ready');
    const [error, setError] = useState<string|null>(null);

    const isError = () => {return state == 'error';}
    const isLoading = () => {return state == 'loading';}
    const isReady = () => {return state == 'ready';}

    const search = async (searchVal: string, orderId: number|null = null) => {
        // reset state for new job number
        setState('loading');
        setError(null);

        // list all potential errors
        const errorKeys: Readonly<{
            [errorKey in ErrorKey]: string;
        }> = {
            invalidJobNumber: t`Invalid job number. Please enter a valid job number.`,
            noResultsForJob: t`Job not found` + ': ' + searchVal,
            cannotFetchData: t`Could not fetch data for job` + ': ' + searchVal,
            unknownError: t`An unknown error has occurred. Please try again.`
        };
        
        // get job info
        const jobInfo = await getJobByNumber(searchVal, orderId);
        if (jobInfo && jobInfo.error) {
            setState('error');
            let errorKey: ErrorKey | null = null;
            if (!['invalidJobNumber', 'noResultsForJob', 'cannotFetchData'].includes(jobInfo.error)) {
                errorKey = 'unknownError';
            } else {
                errorKey = jobInfo.error;
            }
            if (!errorKey) throw new Error('Empty errorKey');
            const errorString = errorKeys[errorKey];
            setError(errorString);
            sendEnqueueSnackbar(t`Failed to find job history.`, {variant: "error"}, enqueueSnackbar);
            return;
        }

        setJobInfo(jobInfo);
        setJobNumber(searchVal);
        setOrderValue(jobInfo.selected_order);
        setState('ready');
        sendEnqueueSnackbar(t`Job history found.`, {variant: "success"}, enqueueSnackbar);
    };

    const orderOptions = () => {
        if (!jobInfo) return [];
        const mapped = jobInfo.all_orders_update_times.map((info) => {
            const systemSettingModel = SystemSettingStore.selectedModel as unknown as SystemSettingModel;
            return {label: systemSettingModel.printDateTime(new Date(info.created_at)), value: info.order_id}
        });
        return mapped;
    }

    const orderSelectOnChange = (orderId: number) => {
        if(!jobInfo) return;
        setOrderValue(orderId);
        search(jobNumber as string, orderId);
    }

    return (<WorkstationHelper stores={[]}>
        <Grid container>
            <Grid item className={classes.searchBoxFull}>
                <OperatorScannerField searchFieldPlaceHolderText={t`Job`} searchFunction={search} externalLoading={isLoading()}/>
            </Grid>
        </Grid>
        {!isError() && jobInfo && <WorkstationGenericWrapper>
            <div className={classes.results}>
                <Paper>
                    <div className={classes.chip}>
                        <Typography variant="h4">
                            {t`Viewing results for job` + ': ' + jobNumber}
                        </Typography>
                        <Typography variant="h6" className={classes.createdAtText}>
                            {t`Created at:`}
                        </Typography>
                        <SelectField loading={isLoading()}
                                     onChange={(name, value) => orderSelectOnChange(value as number)}
                                     value={orderValue}
                                     options={orderOptions()}
                                     className={classes.orderSelector}
                                     size={'small'}
                        />
                    </div>
                    <div className={classes.tabsContainer}>
                        <TabsContainer
                            tabs={[
                            {
                                label: t`Machine History`,
                                component: <MachineHistoryTable
                                    machineHistory={jobInfo.machine_history}
                                />,
                                disabled: false,
                                defaultTab: true
                            },
                            {
                                label: t`Job Data`,
                                component: <JobDataTable
                                    jobDataSides={jobInfo.job_data}
                                />,
                                disabled: false,
                                defaultTab: false
                            },
                            {
                                label: t`Job Process Data`,
                                component: <JobProcessDataTable
                                    processData={{
                                        left: getProcessDataForSide(jobInfo.machine_history, Side.LEFT),
                                        right: getProcessDataForSide(jobInfo.machine_history, Side.RIGHT)
                                    }}
                                />,
                                disabled: false,
                                defaultTab: false
                            },
                            {
                                label: t`Routing History`,
                                component: <div>Routing History</div>,
                                disabled: true,
                                defaultTab: false
                            }
                        ]}/>
                    </div>
                </Paper>
            </div>
        </WorkstationGenericWrapper>}
        {isError() && <WorkstationGenericWrapper><WarningBox text={error}/></WorkstationGenericWrapper>}
    </WorkstationHelper>
    );
});
