import * as React from "react";
import {observer} from "mobx-react-lite";
import {
    Conditional,
    FormProps,
    IfFalse,
    IfTrue,
    PaddedPaper,
    sendEnqueueSnackbar,
    WorkstationHelper
} from "@rx-suite/ui";
import {t, Trans} from "@lingui/macro";
import {Button, CircularProgress, Divider, Grid, makeStyles, Typography} from "@material-ui/core";
import {useSnackbar} from "notistack";
import {ProcessBookingWorkstationModel} from "../../stores/ProcessBookingWorkstationStore";
import {ProcessBookingWorkstationJobSearchField} from "../../components/process_booking_workstation/ProcessBookingWorkstationJobSearchField";
import {ProcessBookingWorkstationMachineSearchField} from "../../components/process_booking_workstation/ProcessBookingWorkstationMachineSearchField";
import {MachineStore} from "../../stores/MachineStore";
import Box from "@material-ui/core/Box";

type JobBookingOperatorFormProps = {
    model: ProcessBookingWorkstationModel
} & Omit<FormProps, 'model'>

export type ProcessBookingWorkstationState = 'scan_job' | 'scan_sn' | 'submit' | 'submitting';

export const ProcessBookingWorkstation: React.FunctionComponent<JobBookingOperatorFormProps> = observer(({model}) => {
    const {enqueueSnackbar} = useSnackbar();

    MachineStore.getAll();

    const [state, setState] = React.useState<ProcessBookingWorkstationState>('scan_job');

    const [jobNumber, setJobNumber] = React.useState<string>("");
    const [machineSn, setMachineSn] = React.useState<string>("");

    React.useEffect(() => {
        if(!jobNumber || !machineSn){
            return;
        }

        const machine = availableMachines.find(machine => machine.sn === machineSn);

        if(!machine){
            setState('scan_sn');
            return;
        }

        setState('submit')
    }, [jobNumber, machineSn]);


    const submit = async ()=> {
        setState('submitting')

        try {
            await model.bookProcess(machineSn, jobNumber);

            sendEnqueueSnackbar(`${t`Successfully booked process.`}`, {variant: "success"}, enqueueSnackbar)

            setJobNumber("");
            setMachineSn("");
            setState('scan_job')
        }catch (e) {
            setState('submit')
            sendEnqueueSnackbar(`${t`Error while booking processes:`} ${e}`, {variant: "error"}, enqueueSnackbar)
        }
    }

    const availableMachines = React.useMemo(() => {
        return MachineStore.models.filter(machine => model.machine_ids.includes(machine.id));
    }, [MachineStore.models]);

    return (
        <WorkstationHelper stores={[]}>
            <PaddedPaper style={{height: '400px'}}>
                <Box mb={1}>
                    <Typography variant={'h4'}>{model.name}</Typography>
                    <Divider/>
                </Box>
                <Grid container spacing={2}>
                    <Grid item xs={12}>
                      <ProcessBookingWorkstationJobSearchField
                          jobNumber={jobNumber}
                          setJobNumber={setJobNumber}
                          setWorkstationState={setState}
                          workstationState={state}
                      />
                    </Grid>
                    <Grid item xs={12}>
                        <ProcessBookingWorkstationMachineSearchField
                            machineSn={machineSn}
                            machines={availableMachines}
                            setMachineSn={setMachineSn}
                            workstationState={state}
                        />
                    </Grid>
                    <Grid item>
                        <Conditional condition={state === 'submitting'}>
                            <IfTrue>
                                <CircularProgress size={48}/>
                            </IfTrue>
                            <IfFalse>
                                <Button
                                    type={'submit'}
                                    color={'secondary'}
                                    size={'large'}
                                    onClick={submit}
                                    variant={'contained'}
                                    disabled={state != 'submit'}>
                                    <Trans>Book Process</Trans>
                                </Button>
                            </IfFalse>
                        </Conditional>
                    </Grid>
                </Grid>
            </PaddedPaper>
        </WorkstationHelper>
    )
})