# == Schema Information
#
# Table name: box_coater_events
#
#  id                  :bigint           not null, primary key
#  duration            :decimal(10, 3)
#  event_type          :integer          not null
#  operator_note       :string(4000)
#  recipe              :string(4000)
#  reporting_category  :string(4000)
#  time                :datetime         not null
#  created_at          :datetime         not null
#  updated_at          :datetime         not null
#  box_coater_batch_id :bigint
#  machine_id          :bigint           not null
#
# Indexes
#
#  index_box_coater_events_on_box_coater_batch_id  (box_coater_batch_id)
#  index_box_coater_events_on_machine_id           (machine_id)
#
# Foreign Keys
#
#  fk_rails_...  (box_coater_batch_id => box_coater_batches.id) ON DELETE => nullify
#  fk_rails_...  (machine_id => machines.id) ON DELETE => cascade
#
class BoxCoaterEvent < ApplicationRecord
  include Pushable
  include AssociateByTimeAndMachine

  belongs_to :machine
  has_one :box_coater_batch, autosave: true

  attr_accessor :processing_time

  enum event_type: [:door_open, :door_closed, :engineering_lockout]
  class << self
    def events_for_machines_between(start_time, end_time, machine_ids)
      BoxCoaterEvent.where(machine_id: machine_ids)
                    .where(BoxCoaterEvent.arel_table[:time].gteq(start_time)
                                         .and(BoxCoaterEvent.arel_table[:time].lt(end_time)))
                    .order(time: :asc)
    end

    def last_event_before_time(time, machine_ids)
      BoxCoaterEvent.where(machine_id: machine_ids)
                    .where(BoxCoaterEvent.arel_table[:time].lt(time))
                    .order(time: :desc)
                    .first
    end
  end

  def event_type_to_name
    if is_door_closed?
      'Door Closed'
    elsif is_door_open?
      'Door Open'
    elsif is_engineering_lockout?
      'Engineering Lockout'
    else
      'Unknown'
    end
  end

  def is_door_closed?
    event_type == 'door_closed'
  end

  def is_door_open?
    event_type == 'door_open'
  end

  def is_engineering_lockout?
    event_type == 'engineering_lockout'
  end

  def find_or_create_associated_batch
    return nil if is_door_open? # only door close events have associated batch

    # Find the closest orphan batch and associate it to this event
    batch = find_associated_by_time_and_machine(BoxCoaterBatch, false, {box_coater_event_id: nil})

    # If there's no orphan batch, create a new batch now
    if batch.nil?
      batch = BoxCoaterBatch.new(machine: machine, time: time)
    end

    batch
  end

  def next_in_time
    BoxCoaterBatch.where(machine_id: machine_id).where(BoxCoaterBatch.arel_table[:time].gt(time)).order(time: :asc).limit(1).first
  end

  def previous_in_time
    BoxCoaterBatch.where(machine_id: machine_id).where(BoxCoaterBatch.arel_table[:time].lt(time)).order(time: :desc).limit(1).first
  end

  def to_json(*args)
    render_jbuilder_partial
  end
end
