# == Schema Information
#
# Table name: machines
#
#  id                  :bigint           not null, primary key
#  archived            :boolean          default(FALSE)
#  description         :string(4000)
#  is_conveyor_machine :boolean          default(FALSE)
#  name                :string(4000)
#  push_target_enabled :boolean          default(FALSE)
#  sn                  :string(4000)     not null
#  track_eer_metrics   :boolean          default(FALSE), not null
#  created_at          :datetime         not null
#  updated_at          :datetime         not null
#  machine_type_id     :bigint
#
# Indexes
#
#  index_machines_on_machine_type_id  (machine_type_id)
#  index_machines_on_sn               (sn) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (machine_type_id => machine_types.id)
#
class Machine < ApplicationRecord
  scope :not_archived, -> { where(archived: false) }
  scope :blockers, -> { joins(:machine_type).where({machine_type: {category: 'blocker'}}) }
  include Pushable
  include MachineKpi
  include MachineStateDurationHelpers
  include MachineErrorHelpers
  include BoxCoater::MachineHelper
  include ProductionView
  include RxModule::LatestAssociatedModel

  has_many :machine_states, -> { order(:time) }, foreign_key: "machine_sn", primary_key: :sn
  has_many :problems, foreign_key: "machine_sn", primary_key: :sn, class_name: "Error"
  has_many :events, foreign_key: "machine_sn", primary_key: :sn, class_name: "Event"
  has_many :job_processes
  has_many :jobs, through: :job_processes

  has_many :job_states
  has_many :eer_yields
  has_many :kpi_job_states
  has_many :kpi_routing_kickouts

  has_and_belongs_to_many :machine_groups
  has_and_belongs_to_many :process_booking_workstations

  has_many :machine_machine_rejects
  has_many :machine_rejects, class_name: "Machine", through: :machine_machine_rejects

  has_many :machine_reject_machines, class_name: "MachineMachineReject", foreign_key: :machine_reject_id
  has_many :machines, through: :machine_reject_machines

  has_many :box_coater_events
  has_many :box_coater_batches

  belongs_to :machine_type, required: false

  has_many :machine_machine_data_push_targets
  has_many :machine_data_push_targets, :through => :machine_machine_data_push_targets

  def extended_name
    return sn if name.nil?
    return "#{name} (#{sn})"
  end

  def ensured_name
    name || sn || id
  end

  def current_state
    KpiCurrentMachineState.find_or_create_by(machine_sn: sn)
  end

  def has_machine_type?
    !machine_type_id.nil?
  end

  def belongs_to_group?
    machine_group_ids.size > 0
  end

  #todo: Move to machine state duration KPI helper?
  def set_current_state(machine_state)
    cur_state = current_state
    # If the current machine state is in the future, the machine time sync was off, so replace with newer entry
    if (machine_state.time >= cur_state.time) || (current_state.time >= (Time.zone.now + 15.minutes))
      cur_state.update!(state: machine_state.state, time: machine_state.time, comment: machine_state.comment)
    end
  end

  def self.status_all
    status = {}
    last_machine_state_per_machine = Machine.latest_records_per_foreign_key(MachineState,
                                                            foreign_key_column: :machine_sn,
                                                            base_class_key_column: :sn,
                                                            time_column: :time,
                                                            limit: 1,
                                                            extra_select: [Machine.arel_table[:id].as('machine_id'),
                                                                           Machine.arel_table[:sn],
                                                                           Machine.arel_table[:name],
                                                                           Machine.arel_table[:is_conveyor_machine],
                                                                           Machine.arel_table[:machine_group_ids]

                                                            ],
                                                            extra_where: Machine.arel_table[:archived].eq(0),
                                                            from: add_machine_group_ids_to_machine_query(
                                                              extra_select: [Machine.arel_table[:is_conveyor_machine]],
                                                              extra_group: [Machine.arel_table[:is_conveyor_machine]]
                                                            ))

    last_machine_state_per_machine.each do |machine_machine_state|
      status[machine_machine_state.sn] = {
        machine_id: machine_machine_state.machine_id,
        name: machine_machine_state.name,
        status: DataHelpers::MACHINE_STATES[machine_machine_state.state],
        comment: machine_machine_state.comment,
        time: machine_machine_state.time,
        duration: Time.now - machine_machine_state.time,
        groups: machine_machine_state[:machine_group_ids].nil? ? [] : machine_machine_state[:machine_group_ids].split(","),
        is_conveyor_machine: machine_machine_state.is_conveyor_machine
      }
    end
    status
  end

  def self.last_job_states(limit = 5)
    limit = ensure_limit(limit)
    job_states_per_machine = {}
    job_states = Machine.latest_records_per_foreign_key(JobState,
                                                        limit: limit,
                                                        extra_select: [
                                                          Machine.arel_table[:sn],
                                                          Machine.arel_table[:name],
                                                          Machine.arel_table[:machine_group_ids]
                                                        ],
                                                        extra_where: Machine.arel_table[:archived].eq(0),
                                                        from: add_machine_group_ids_to_machine_query)
    # using each here instead of find_each because the query is too complex for find_each and we are limited to 5 job_states per machine anyways.
    job_states.each do |job_state|
      job_states_per_machine[job_state.sn] ||= []
      attributes = job_state.attributes
      attributes["machine_group_ids"] = attributes["machine_group_ids"].nil? ? [] : attributes["machine_group_ids"].split(',')
      JobState.defined_enums.each do |enum_key, enum_value|
        attributes[enum_key] = enum_value.key(attributes[enum_key])
      end
      job_states_per_machine[job_state.sn] << attributes
    end
    job_states_per_machine
  end

  def self.ensure_limit(limit)
    limit = limit.nil? ? 5 : limit
    limit = limit.to_i > 1000 ? 1000 : limit.to_i
    limit
  end

  def self.machine_list
    options = []
    other = {label: 'other', options: []}
    Machine.find_each do |machine|
      if machine.has_machine_type?
        type = options.find {|c| c[:label] == machine.machine_type.category}
        if type.nil?
          type = {label: machine.machine_type.category, options: []}
          options << type
        end
        type[:options] << { label: machine.extended_name, value: machine.id }
      else
        other[:options] << { label: machine.extended_name, value: machine.id }
      end
    end
    options << other
    options
  end

  def self.add_machine_group_ids_to_machine_query(extra_select: [], extra_group: [])
    # This query's purpose is to add the column machine_group_ids to the machine records
    cast_type = if DatabaseServer.mssql?
                  'nvarchar'
                else
                  'varchar'
                end
    # Adding machine group ids as a column to the resulting query
    machine_groups_machines_table = Arel::Table.new(:machine_groups_machines)
    Machine.select(Machine.arel_table[:id],
                   Machine.arel_table[:name],
                   Machine.arel_table[:sn],
                   Machine.arel_table[:archived],
                   Arel::Nodes::NamedFunction.new('STRING_AGG',
                                                  [MachineGroup.arel_table[:id].cast(cast_type),
                                                   Arel::Nodes.build_quoted(',')])
                                             .as('machine_group_ids'),
                   *extra_select)
           .joins(Machine.arel_table
                         .join(machine_groups_machines_table, Arel::Nodes::OuterJoin)
                         .on(Machine.arel_table[:id].eq(machine_groups_machines_table[:machine_id]))
                         .join_sources)
           .joins(Machine.arel_table
                         .join(MachineGroup.arel_table, Arel::Nodes::OuterJoin)
                         .on(machine_groups_machines_table[:machine_group_id].eq(MachineGroup.arel_table[:id]))
                         .join_sources)
           .group(Machine.arel_table[:id],
                  Machine.arel_table[:name],
                  Machine.arel_table[:sn],
                  Machine.arel_table[:archived],
                  *extra_group)
  end

  def push_targets_for_path(path)
    MachineDataPushTarget.where(send_to: :all_machines).where.not(path => [nil, ""]) + self.machine_data_push_targets.where.not(path => [nil, ""])
  end

  # TODO: this should be retired once 'current_state' is fully implemented
  def last_state
    @last_state ||= machine_states.last
  end

  #TODO: delete?
  reportable("machine", "Machine", "Report on machines", {
    name: {
      name: "Name",
      desc: "The name of the machine"
    },
    sn: {
      name: "Serial Number",
      shortname: "SN",
      desc: "The serial number of the machine"
    },
    description: {
      name: "Description",
      desc: "The description of the machine"
    },
    is_conveyor_machine: {
      name: "Conveyor?",
      desc: "Whether or not the machine is a conveyor"
    },
  })
end
