require "socket"
require "singleton"

class BlockReuseServer
  include Singleton

  def initialize
    @server = nil
    @listener_thread = nil
    @clients = Set.new
  end

  def start(port)
    unless @server.nil?
      Rails.logger.warn "ReUse Server is already started."
      return
    end

    BlockReuseServer.log "Starting server on port #{port}"

    @server = TCPServer.new("0.0.0.0", port)
    @listener_thread = Thread.new do
      Rails.application.executor.wrap do
        loop do
          # Stop listening for new connections if the server is stopped
          break if @server.nil?
          @clients << Thread.new(@server.accept) do |client|
            BlockReuseServer.log "incoming connection from #{request_ip(client)}"

            Rails.application.executor.wrap do
              process_reuse_request(client)
              begin
                client.close
              rescue => err
                BlockReuseServer.log "Failed to close socket to client - #{err}"
              end
              @clients.delete Thread.current
            end
          end
        end
      end
    end
  end

  def stop
    return if @server.nil?

    BlockReuseServer.log "Stopping server"

    BlockReuseServer.log "Waiting for #{@clients.count} processes to finish..."

    wait_for_clients_to_finish

    BlockReuseServer.log "Finished processing connections"

    kill_accept_loop
    close_server

    BlockReuseServer.log "Server stopped"
  end

  def online?
    @server.present?
  end

  private

  def self.log(msg)
    Rails.logger.info "[ReUse Server] #{msg}"
  end

  def wait_for_clients_to_finish
    @clients.each(&:join)
  end

  def kill_accept_loop
    @listener_thread&.kill
    @listener_thread = nil
  end

  def close_server
    @server&.close
    @server = nil
  end

  def request_ip(client)
    client.peeraddr.nil? ? "unknown" : client.peeraddr[3]
  end

  def process_reuse_request(client)
    begin
      # Header is always 64 bytes and sent as packed little endian bytes
      # header_raw = client.recv(64).bytes
      header_raw = client.recv(64)
      req_header = unpack_header(header_raw)

      BlockReuseServer.log "request headers -> #{req_header}"

      # The error and body data are not encoded and can be used as-is

      if req_header[:error_size] > 0
        # We don't do anything with this yet but we still need to receive that data for us to correctly receive the body
        _ = client.recv(req_header[:error_size])
      end

      if req_header[:data_size] <= 0
        reply_invalid_reuse_request(req_header, "Request header has no data.", client)
      end

      req_body = get_body(client, req_header)
      BlockReuseServer.log "Command #{req_header[:command]} specified"
      case req_header[:command]
        when 1100
          process_command_1100(client, req_header, req_body)
        when 1103
          process_command_1103(client, req_header, req_body)
        when 1107
          process_command_1107(client, req_header, req_body)
        else
          reply_invalid_reuse_request(req_header, "Invalid command.\n", client)
      end
    rescue => err
      BlockReuseServer.log "Error while processing request from #{request_ip(client)} - #{err}"
    end
  end

  def reply_invalid_reuse_request(req_header, error, client)
    BlockReuseServer.log "error -> #{error}"

    # Any packet sent must end in new line for the client to accept it
    # While we could add the newline to the error message, it would result
    # in the UI displaying weird characters to the user dialog.
    res_body = "\n"

    res_header = pack_header(
      {
        sender: req_header[:sender],
        command: req_header[:command],
        status: 4,
        error: 1006,
        error_size: error.length,
        mid: req_header[:mid],
        version: req_header[:version],
        data_type: req_header[:data_type],
        data_version: req_header[:data_version],
        data_size: res_body.length
      }
    )

    client.puts(res_header + error + res_body)
  end

  def reply_block_code(req_header, req_body, block, client, command)
    res_body = "_BLKC=#{block.dmc}\n" \
      "_BLKCV=V1\n"

    res_header = pack_header(
      {
        sender: req_header[:sender],
        command: command,
        status: 2,
        error: 0,
        error_size: 0,
        mid: req_header[:mid],
        version: req_header[:version],
        data_type: req_header[:data_type],
        data_version: req_header[:data_version],
        data_size: res_body.length
      }
    )

    client.puts(res_header + res_body)
  end

  def reply_block_data(req_header, req_body, block, client, command)
    dmc_data = parse_block_dmc block.dmc

    block_reuse_code = 1
    block_reuse_code_value = "ReUse"
    if block.cycles_until_wash <= 0
      block_reuse_code = 2
      block_reuse_code_value = "Wash"
    end

    # Discard has priority over the other states
    if block.cycles_until_discard <= 0
      block_reuse_code = 0
      block_reuse_code_value = "Discard"
    end



    # Blocks imported from the old ReUse server may not have job associated with them
    job_number = block.job.nil? ? 0 : block.job.number
    lens_side = block.side.nil? ? :left : block.side

    res_body = "JOB=#{job_number}\n" \
      "_BLKLS=#{Sideable.side_to_side_num(lens_side)}\n" \
      "_BLKC=#{block.dmc}\n" \
      "_BLKCV=1\n" \
      "_BLKKNW=1\n" \
      "_BLKID=#{dmc_data[:block_id]}\n" \
      "_BLKMNU=#{dmc_data[:manufacturer]}\n" \
      "_BLKPDATE=#{dmc_data[:prod_date]}\n" \
      "_BLKSH=3\n" \
      "_BLKDM1=#{format_precision(dmc_data[:diameter])}\n" \
      "_BLKRD=#{format_precision(dmc_data[:radius])}\n" \
      "_BLKCEHI=#{format_precision(format_block_height(dmc_data[:height]))}\n" \
      "_BLKDECX=#{format_precision("0")}\n" \
      "_BLKDECY=#{format_precision("0")}\n" \
      "_BLKRU=#{block_reuse_code}\n" \
      "_BLKCYCLES=#{block.usage_count}\n"

    BlockReuseServer.log "Replying to command #{command} from #{request_ip(client)} with code #{block_reuse_code}(#{block_reuse_code_value}). response body: #{res_body}"

    res_header = pack_header(
      {
        sender: req_header[:sender],
        command: command,
        status: 2,
        error: 0,
        error_size: 0,
        mid: req_header[:mid],
        version: req_header[:version],
        data_type: req_header[:data_type],
        data_version: req_header[:data_version],
        data_size: res_body.length
      }
    )

    client.puts(res_header + res_body)
  end

  def parse_request_body(body_raw)
    data = body_raw.split("\n")

    job_number_tag = data.grep(/JOB/)
    lens_side_tag = data.grep(/_BLKLS/)
    dmc = data.grep(/_BLKC/)

    {
      job_number: job_number_tag.empty? ? nil : extract_tag_value(job_number_tag[0]),
      lens_side: lens_side_tag.empty? ? nil : extract_tag_value(lens_side_tag[0]),
      dmc: dmc.empty? ? nil : extract_tag_value(dmc[0])
    }
  end

  def get_body(client, req_header)
    body_raw = client.recv(req_header[:data_size])
    req_body = parse_request_body(body_raw)
    BlockReuseServer.log "Request body -> #{req_body}"

    # Attempt to save the message for trace purposes
    reuse_block_settings = BlockReuseSetting.first
    unless reuse_block_settings.nil?
      reuse_block_settings.last_server_message = req_body
      reuse_block_settings.save
    end
    req_body
  end

  def parse_block_dmc(dmc)
    {
      radius: dmc[0..2],
      diameter: dmc[3..4],
      height: dmc[5..7],
      type: dmc[8],
      manufacturer: dmc[9],
      prod_date: dmc[10..15],
      block_id: dmc[16..21]
    }
  end

  def format_precision(value)
    '%.3f'%value
  end

  # Block height has to be returned with a decimal precision of 1 (ex: 6.0)
  def format_block_height(height)
    # If the height has a lead 0, we have to remove the 0 otherwise it will be parsed as Octal
    height = height.sub(/^0+/, "") if height[0] == "0"
    height.to_f/10
  end

  def extract_tag_value(tag)
    tag.split("=").last.split("\n").first.chomp
  end

  def unpack_header(bytes)
    # The header is 64 bytes sent in a 4 byte little endian packed format, so process the data in groups of 4 bytes
    data = bytes.unpack("V16")
    {
      sender: data[0],
      command: data[1],
      status: data[2],
      error: data[3],
      error_size: data[4],
      mid: data[5],
      version: data[6],
      data_type: data[7],
      data_version: data[8],
      data_size: data[9],
      # The header data only makes up for the first 40 bytes, the rest is padded with a reserved area for now
      reserved: data[10..15]
    }
  end

  def pack_header(header)
    data = header.values

    # Pad for the reserved area (24 bytes)
    6.times do
      data << 0
    end

    # return the data in the same way it was formatted, groups of 4 bytes in little endian format
    data.pack("V16")
  end

  # Command 1100: Request block data using the Block's DMC.
  def process_command_1100(client, req_header, req_body)
    if req_body[:dmc].nil?
      return reply_invalid_reuse_request(req_header, "Request body is missing data.", client)
    end
    block = Block.find_by(dmc: req_body[:dmc])
    if block.nil?
      return reply_invalid_reuse_request(req_header, "No block found from the dmc.", client)
    end
    reply_block_data(req_header, req_body, block, client, 1100)
  end

  # Command 1103: Request block data using the last job number+lens side.
  def process_command_1103(client, req_header, req_body)
    if req_body[:job_number].nil? || req_body[:lens_side].nil?
      return reply_invalid_reuse_request(req_header, "Request body is missing data.", client)
    end

    job = FindHelpers::JobFinders.latest_with_number_and_side_num(req_body[:job_number], req_body[:lens_side])

    if job.nil?
      return reply_invalid_reuse_request(req_header, "No job with number #{req_body[:job_number]} [side: #{req_body[:lens_side]}] was found.", client)
    end

    block = Block.where(job_id: job.id).first

    if block.nil?
      return reply_invalid_reuse_request(req_header, "No block is currently attached to the requested job.", client)
    end

    reply_block_data(req_header, req_body, block, client, 1103)
  end

  # Command 1107: Request that the block specified by the DMC is washed.
  def process_command_1107(client, req_header, req_body)
    if req_body[:dmc].nil?
      return reply_invalid_reuse_request(req_header, "Request body is missing data.", client)
    end

    block = Block.find_by(dmc: req_body[:dmc])
    if block.nil?
      return reply_invalid_reuse_request(req_header, "No block found from the dmc.", client)
    end

    # Wash block
    settings = BlockReuseSetting.first

    if settings.nil?
      return reply_invalid_reuse_request(req_header, "No block reuse settings set.", client)
    end

    block.cycles_until_wash = settings.block_wash_after
    block.save

    reply_block_code(req_header, req_body, block, client, 1107)
  end
end
